/*
    Copyright  1995-2001, The AROS Development Team. All rights reserved.
    $Id: datatypes_init.c,v 1.3 2012/03/27 22:21:53 piru Exp $

    Desc: initialize datatypes.library
    Lang: English.
*/

#include <exec/types.h>
#include <exec/lists.h>
#include <exec/memory.h>
#include <exec/resident.h>
#include <aros/libcall.h>
#include <exec/semaphores.h>
#include <aros/asmcall.h>

#include <proto/exec.h>

#include <aros/debug.h>

#include "datatypes_intern.h"
#include "libdefs.h"

#define INIT AROS_SLIB_ENTRY(init, DataTypes)
static const char name[];
static const char version[];
static const APTR inittabl[4];
static const void * const LIBFUNCTABLE[];
struct DataTypesBase * INIT();
extern const char LIBEND;

int entry(void)
{
    return -1;
}

const struct Resident DataTypes_resident =
{
    RTC_MATCHWORD,
    (struct Resident *)&DataTypes_resident,
    (APTR)&LIBEND,
#ifdef __MORPHOS__
    RTF_PPC | RTF_EXTENDED | RTF_AUTOINIT,
#else
    RTF_AUTOINIT,
#endif
    VERSION_NUMBER,
    NT_LIBRARY,
    -120,
    (char *)name,
    (char *)&version[6],
    (ULONG *)inittabl,
#ifdef __MORPHOS__
    REVISION_NUMBER,	/* Revision */
    NULL /* Tags */
#endif
};

static const char name[]=NAME_STRING;
static const char version[]=VERSION_STRING;

static const APTR inittabl[4] =
{
    (APTR)sizeof(struct DataTypesBase),
    (APTR)LIBFUNCTABLE,
    NULL,
    &INIT
};

static void closelibs(struct DataTypesBase *DataTypesBase)
{
    if(DataTypesBase->dtb_IntuitionBase != NULL)
	CloseLibrary(DataTypesBase->dtb_IntuitionBase);

    if(DataTypesBase->dtb_LayersBase != NULL)
	CloseLibrary(DataTypesBase->dtb_LayersBase);

    if(DataTypesBase->dtb_UtilityBase != NULL)
	CloseLibrary(DataTypesBase->dtb_UtilityBase);

    if(DataTypesBase->dtb_DOSBase != NULL)
	CloseLibrary(DataTypesBase->dtb_DOSBase);

    if(DataTypesBase->dtb_IFFParseBase != NULL)
	CloseLibrary(DataTypesBase->dtb_IFFParseBase);

    if(DataTypesBase->dtb_LocaleBase != NULL)
	CloseLibrary(DataTypesBase->dtb_LocaleBase);

    if(DataTypesBase->dtb_GfxBase != NULL)
	CloseLibrary(DataTypesBase->dtb_GfxBase);

    if(DataTypesBase->dtb_IconBase != NULL)
	CloseLibrary(DataTypesBase->dtb_IconBase);

    if (DataTypesBase->dtb_WorkbenchBase != NULL)
	CloseLibrary(DataTypesBase->dtb_WorkbenchBase);

    if(DataTypesBase->dtb_RexxSysBase != NULL)
	CloseLibrary(DataTypesBase->dtb_RexxSysBase);

    FreeMem(((UBYTE *) DataTypesBase) - DataTypesBase->dtb_LibNode.lib_NegSize,
            DataTypesBase->dtb_LibNode.lib_PosSize + DataTypesBase->dtb_LibNode.lib_NegSize);
}


#ifdef __MORPHOS__
struct DataTypesBase *LIB_init(struct DataTypesBase *LIBBASE, BPTR segList, struct ExecBase *sysBase)
{
#else
AROS_UFH3(struct DataTypesBase *, AROS_SLIB_ENTRY(init,DataTypes),
    AROS_UFHA(struct DataTypesBase *,  DataTypesBase, D0),
    AROS_UFHA(BPTR,                  segList, A0),
    AROS_UFHA(struct ExecBase *, sysBase, A6)
)
#endif
{
    AROS_USERFUNC_INIT
    int i;

    DataTypesBase->dtb_SysBase = sysBase;
#undef SysBase
    SysBase = (struct ExecBase *)sysBase;
#define SysBase ((struct DataTypesBase *)DataTypesBase)->dtb_SysBase
    DataTypesBase->dtb_SegList = segList;
    
    for (i = 0; i < SEM_MAX; i++)
    {
	InitSemaphore(&DataTypesBase->dtb_Semaphores[i]);
    }

    /*
     * Open libraries. These should all exist at init time, so there is no
     * point not opening them now. If they happen to fail, we do not load
     * the library. Under debugging mode print out why though. I don't
     * alert because a) datatypes isn't that important, and b) it may be a
     * version problem. In either case I don't want to bring down the
     * system unnecessarily.
     */

    #warning "FIXME: Opening disk based libraries breaks forbid! - Piru"

    if ((DataTypesBase->dtb_UtilityBase =
	    OpenLibrary("utility.library", 39L)) == NULL)
    {
	D(bug("datatypes.library: Cannot open utility.library\n"));
	goto error;
    }

    if ((DataTypesBase->dtb_GfxBase =
	    OpenLibrary("graphics.library", 39L)) == NULL)
    {
	D(bug("datatypes.library: Cannot open graphics.library\n"));
	goto error;
    }

    if ((DataTypesBase->dtb_LayersBase = 
	    OpenLibrary("layers.library", 39L)) == NULL)
    {
	D(bug("datatypes.library: Cannot open layers.library\n"));
	goto error;
    }

    if ((DataTypesBase->dtb_IntuitionBase =
	    OpenLibrary("intuition.library", 39L)) == NULL)
    {
	D(bug("datatypes.library: Cannot open intuition.library\n"));
	goto error;
    }

    if ((DataTypesBase->dtb_DOSBase =
	    OpenLibrary("dos.library", 37L)) == NULL)
    {
	D(bug("datatypes.library: Cannot open dos.library\n"));
	goto error;
    }

    /* We may not have these libraries, but try anyway */
    if ((DataTypesBase->dtb_IFFParseBase =
	    OpenLibrary("iffparse.library", 37L)) == NULL)
    {
	D(bug("datatypes.library: Cannot open iffparse.library\n"));
	goto error;
    }
#if 0
    if ((DataTypesBase->dtb_LocaleBase =
	    OpenLibrary("locale.library", 0L)) == NULL)
    {
	D(bug("datatypes.library: Cannot open locale.library\n"));
	goto error;
    }
#endif

    if ((DataTypesBase->dtb_IconBase =
	    OpenLibrary("icon.library", 37L)) == NULL)
    {
	D(bug("datatypes.library: Cannot open icon.library\n"));
	goto error;
    }

    if ((DataTypesBase->dtb_WorkbenchBase =
	    OpenLibrary("workbench.library", 37L)) == NULL)
    {
	D(bug("datatypes.library: Cannot open workbench.library\n"));
	goto error;
    }

#if 1
    /* Open locale library, if we can't open it, just don't use it! - Piru */
    DataTypesBase->dtb_LocaleBase = OpenLibrary("locale.library", 37L);
#endif

    /* If it doesn't open, we won't use it */
    DataTypesBase->dtb_RexxSysBase = OpenLibrary("rexxsyslib.library", 36L);

    /* Get the list of datatypes */
    DataTypesBase->dtb_DTList = GetDataTypesList(DataTypesBase);

    if(!InstallClass((struct Library *)DataTypesBase))
    {
	//return NULL;
	D(bug("datatypes.library: Cannot InstallClass\n"));
	goto error;
    }

    /* Try opening the catalog, don't worry if we fail, just keep trying. */
    DataTypesBase->dtb_LibsCatalog = opencatalog(
	    (struct Library *)DataTypesBase,
	    NULL,
	    "Sys/libs.catalog",
	    OC_BuiltInLanguage,
	    "english",
	    TAG_DONE);

    DataTypesBase->dtb_LibNode.lib_Revision = REVISION_NUMBER;

    return DataTypesBase;

error:
    closelibs(DataTypesBase);

    return NULL;

    AROS_USERFUNC_EXIT
}


AROS_LH1(struct DataTypesBase *, open,
    AROS_LHA(ULONG, version, D0),
    struct DataTypesBase *, DataTypesBase, 1, DataTypes)
{
    AROS_LIBFUNC_INIT

    /* Keep the compiler happy */
    (void) version;

    /* What else do we have to do? */
    Forbid();
    DataTypesBase->dtb_LibNode.lib_OpenCnt++;
    DataTypesBase->dtb_LibNode.lib_Flags &= ~LIBF_DELEXP;
    Permit();

    /* Try opening the catalog again, if not open already. */
    if (!DataTypesBase->dtb_LibsCatalog)
    {
	DataTypesBase->dtb_LibsCatalog =
	    opencatalog
	    (
		(struct Library *)DataTypesBase,
		NULL,
		"Sys/libs.catalog",
		OC_BuiltInLanguage,
		"english",
		TAG_DONE
	    );
    }

    return DataTypesBase;
    AROS_LIBFUNC_EXIT
}


AROS_LH0(BPTR, close,
    struct DataTypesBase *, DataTypesBase, 2, DataTypes)
{
    AROS_LIBFUNC_INIT

    Forbid();

    --DataTypesBase->dtb_LibNode.lib_OpenCnt;

    if (DataTypesBase->dtb_LibNode.lib_OpenCnt == 0 &&
        (DataTypesBase->dtb_LibNode.lib_Flags & LIBF_DELEXP))
    {
	/*
	    We can never exit because of the system patches,
	    But we can try and free some memory.
	*/

#if USE_MEMFS
	memfs_stop((struct Library *) DataTypesBase);
#endif

	AROS_LC0(BPTR, expunge, LIBBASETYPE *, DataTypesBase, 3, DataTypes);
    }

    Permit();

    return 0;
    AROS_LIBFUNC_EXIT
}


AROS_LH0(BPTR, expunge,
    struct DataTypesBase *, DataTypesBase, 3, DataTypes)
{
    AROS_LIBFUNC_INIT

    DataTypesBase->dtb_LibNode.lib_Flags |= LIBF_DELEXP;

    /* As I said above, we cannot remove ourselves. */

#if 0
    /* stegerg: if later someone else re-opens datatypes.library, then
                the datatypes.class would have to be re-added with
		AddClass in libopen() (if FreeClass returned FALSE,
		where the class was not freed, but still removed),
		or re-make the class (when FreeClass returned TRUE) */
		
    TryRemoveClass((struct Library *)DataTypesBase);
#endif

    return 0;
    AROS_LIBFUNC_EXIT
}


AROS_LH0I(int, null,
    struct DataTypesBase *, LIBBASE, 4, DataTypes)
{
    AROS_LIBFUNC_INIT

    return 0;

    AROS_LIBFUNC_EXIT
}

